from .game_objects import GameObject, NoObject
from .utils import find_objects, match_objects, match_blinking_objects
import numpy as np
from scipy.ndimage import label, find_objects as nd_find_objects, binary_closing

objects_colors = {"yellow": [252, 252, 84], "green": [135, 183, 84], "red": [214, 92, 92], "black": [0, 0, 0]}

class Player(GameObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rgb = [252, 252, 84]


class Warrior(GameObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rgb = [135, 183, 84]
        self.num_frames_invisible = -1
        self.max_frames_invisible = 4


class Pig(GameObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rgb = [214, 92, 92]
        self.num_frames_invisible = -1
        self.max_frames_invisible = 4

class Shadow(GameObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rgb = [1, 1, 1]
        self.num_frames_invisible = -1
        self.max_frames_invisible = 4


class Chicken(GameObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rgb = [252, 252, 84]
        self.num_frames_invisible = -1
        self.max_frames_invisible = 4

#  ---- HUD -----
class Score(GameObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rgb = [252, 252, 84]
        self.num_frames_invisible = -1
        self.max_frames_invisible = 4


class Life(GameObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rgb = [252, 252, 84]
        self.num_frames_invisible = -1
        self.max_frames_invisible = 4


def _detect_objects(objects, obs, hud=False):
    global path_pixels
    warrior_bb = []
    pig_bb = []
    shadow_bb = []
    chicken_bb = []
    enemy_type = Warrior
    player = objects[0]

    yellow = find_objects(obs, objects_colors["yellow"], maxy=170)
    for bb in yellow:
        if bb[2] > 6:
            player.xy = bb[:2]
        else:
            chicken_bb.append(bb)

    green = find_objects(obs, objects_colors["green"])
    for bb in green:
        warrior_bb.append(bb)

    red = find_objects(obs, objects_colors["red"])
    for bb in red:
        pig_bb.append(bb)

    shadows = find_shadows(obs, objects_colors["black"], closing_dist=3)
    for bb in shadows:
        shadow_bb.append(bb)

    # Used to get the list of path pixels -- not relevant anymore unless something is broken
    # path_pixels = accumulate_path_pixels(obs, path_pixels=path_pixels)
    # with open("path_pixels.txt", "w") as file:
    #     for i, pixel in enumerate(path_pixels):
    #         file.write(f"{pixel}, ")
    #         if (i%30==0):
    #             file.write(f"\n")

    match_blinking_objects(objects, warrior_bb, 1, 6, Warrior)
    
    match_blinking_objects(objects, pig_bb, 7, 6, Pig)
    
    match_blinking_objects(objects, shadow_bb, 13, 6, Shadow)

    match_blinking_objects(objects, chicken_bb, 19, 6, Chicken)

    if hud:
        score = objects[-4]
        lives_bbs = find_objects(obs, objects_colors["yellow"], miny=175, closing_dist=4, min_distance=1)
        for bb in lives_bbs:
            if bb[2] > 4:
                score.xywh = bb
                lives_bbs.remove(bb)
                break
        match_objects(objects, lives_bbs, -3, 3, Life)

def find_shadows(current_frame, shadow_color, closing_dist=1):
    """
    Find shadows by checking if any of the path pixels are the shadow color and merge close shadows.
    
    Parameters:
    - current_frame: The current observation image.
    - shadow_color: The color of the shadows.
    - closing_dist: The distance within which close shadows are merged.
    
    Returns:
    - A list of bounding boxes for the detected shadows.
    """
    detected_shadows = []
    height, width, _ = current_frame.shape
    shadow_color = np.array(shadow_color)
    
    shadow_mask = np.zeros((height, width), dtype=bool)
    
    for x, y in path_pixels:
        current_pixel = current_frame[y, x]
        
        if np.all(np.abs(current_pixel - shadow_color) == 0):
            shadow_mask[y, x] = True
    
    # Apply binary closing to merge close shadows
    structure = np.ones((closing_dist, closing_dist), dtype=bool)
    closed_shadow_mask = binary_closing(shadow_mask, structure=structure)
    
    # Label connected components
    labeled_array, num_features = label(closed_shadow_mask)
    
    # Find bounding boxes of connected components
    objects_slices = nd_find_objects(labeled_array)
    
    for obj_slice in objects_slices:
        y_slice, x_slice = obj_slice
        x_min, x_max = x_slice.start, x_slice.stop
        y_min, y_max = y_slice.start, y_slice.stop
        detected_shadows.append((x_min, y_min, x_max - x_min, y_max - y_min))
    
    return detected_shadows


def accumulate_path_pixels(frame, path_pixels, path_color=[162, 98, 33]):
    """
    Used to get the list of pixels that need to be checked for shadows.
    Accumulate path pixels over time.
    
    Parameters:
    - frame: The current observation image.
    - path_color: The color of the path stripes.
    
    Returns:
    - A list of path pixels.
    """
    path_color = np.array(path_color)
    height, width, _ = frame.shape
    new_path_pixels = []
    
    for y in range(height):
        for x in range(width):
            pixel = frame[y, x]
            if np.all(np.abs(pixel - path_color) == 0):
                new_path_pixels.append((x, y))
    
    return sorted(list(set(new_path_pixels + path_pixels)))

path_pixels = [(16, 15), (16, 17), (16, 18), (16, 20), (16, 22), (16, 23), (16, 25), (16, 27), (16, 28), (16, 30), (16, 32), (16, 33), (16, 35), (16, 37), (16, 38), (16, 40), (16, 42), (16, 43), (16, 45), (16, 47), (16, 48), (16, 50), (16, 52), (16, 53), (16, 55), (16, 57), (16, 58), (16, 60), (16, 62), (16, 63), (16, 65), 
(16, 67), (16, 68), (16, 70), (16, 72), (16, 73), (16, 75), (16, 77), (16, 78), (16, 80), (16, 82), (16, 83), (16, 85), (16, 87), (16, 88), (16, 90), (16, 92), (16, 93), (16, 95), (16, 97), (16, 98), (16, 100), (16, 102), (16, 103), (16, 105), (16, 107), (16, 108), (16, 110), (16, 112), (16, 113), (16, 115), 
(16, 117), (16, 118), (16, 120), (16, 122), (16, 123), (16, 125), (16, 127), (16, 128), (16, 130), (16, 132), (16, 133), (16, 135), (16, 137), (16, 138), (16, 140), (16, 142), (16, 143), (16, 145), (16, 147), (16, 148), (16, 150), (16, 152), (16, 153), (16, 155), (16, 157), (16, 158), (16, 160), (16, 162), (16, 163), (16, 165), 
(16, 167), (17, 15), (17, 17), (17, 18), (17, 20), (17, 22), (17, 23), (17, 25), (17, 27), (17, 28), (17, 30), (17, 32), (17, 33), (17, 35), (17, 37), (17, 38), (17, 40), (17, 42), (17, 43), (17, 45), (17, 47), (17, 48), (17, 50), (17, 52), (17, 53), (17, 55), (17, 57), (17, 58), (17, 60), (17, 62), 
(17, 63), (17, 65), (17, 67), (17, 68), (17, 70), (17, 72), (17, 73), (17, 75), (17, 77), (17, 78), (17, 80), (17, 82), (17, 83), (17, 85), (17, 87), (17, 88), (17, 90), (17, 92), (17, 93), (17, 95), (17, 97), (17, 98), (17, 100), (17, 102), (17, 103), (17, 105), (17, 107), (17, 108), (17, 110), (17, 112), 
(17, 113), (17, 115), (17, 117), (17, 118), (17, 120), (17, 122), (17, 123), (17, 125), (17, 127), (17, 128), (17, 130), (17, 132), (17, 133), (17, 135), (17, 137), (17, 138), (17, 140), (17, 142), (17, 143), (17, 145), (17, 147), (17, 148), (17, 150), (17, 152), (17, 153), (17, 155), (17, 157), (17, 158), (17, 160), (17, 162), 
(17, 163), (17, 165), (17, 167), (18, 15), (18, 17), (18, 18), (18, 20), (18, 22), (18, 23), (18, 25), (18, 27), (18, 28), (18, 30), (18, 32), (18, 33), (18, 35), (18, 37), (18, 38), (18, 40), (18, 42), (18, 43), (18, 45), (18, 47), (18, 48), (18, 50), (18, 52), (18, 53), (18, 55), (18, 57), (18, 58), 
(18, 60), (18, 62), (18, 63), (18, 65), (18, 67), (18, 68), (18, 70), (18, 72), (18, 73), (18, 75), (18, 77), (18, 78), (18, 80), (18, 82), (18, 83), (18, 85), (18, 87), (18, 88), (18, 90), (18, 92), (18, 93), (18, 95), (18, 97), (18, 98), (18, 100), (18, 102), (18, 103), (18, 105), (18, 107), (18, 108), 
(18, 110), (18, 112), (18, 113), (18, 115), (18, 117), (18, 118), (18, 120), (18, 122), (18, 123), (18, 125), (18, 127), (18, 128), (18, 130), (18, 132), (18, 133), (18, 135), (18, 137), (18, 138), (18, 140), (18, 142), (18, 143), (18, 145), (18, 147), (18, 148), (18, 150), (18, 152), (18, 153), (18, 155), (18, 157), (18, 158), 
(18, 160), (18, 162), (18, 163), (18, 165), (18, 167), (19, 15), (19, 17), (19, 18), (19, 20), (19, 22), (19, 23), (19, 25), (19, 27), (19, 28), (19, 30), (19, 32), (19, 33), (19, 35), (19, 37), (19, 38), (19, 40), (19, 42), (19, 43), (19, 45), (19, 47), (19, 48), (19, 50), (19, 52), (19, 53), (19, 55), 
(19, 57), (19, 58), (19, 60), (19, 62), (19, 63), (19, 65), (19, 67), (19, 68), (19, 70), (19, 72), (19, 73), (19, 75), (19, 77), (19, 78), (19, 80), (19, 82), (19, 83), (19, 85), (19, 87), (19, 88), (19, 90), (19, 92), (19, 93), (19, 95), (19, 97), (19, 98), (19, 100), (19, 102), (19, 103), (19, 105), 
(19, 107), (19, 108), (19, 110), (19, 112), (19, 113), (19, 115), (19, 117), (19, 118), (19, 120), (19, 122), (19, 123), (19, 125), (19, 127), (19, 128), (19, 130), (19, 132), (19, 133), (19, 135), (19, 137), (19, 138), (19, 140), (19, 142), (19, 143), (19, 145), (19, 147), (19, 148), (19, 150), (19, 152), (19, 153), (19, 155), 
(19, 157), (19, 158), (19, 160), (19, 162), (19, 163), (19, 165), (19, 167), (20, 15), (20, 17), (20, 18), (20, 45), (20, 47), (20, 48), (20, 75), (20, 77), (20, 78), (20, 105), (20, 107), (20, 108), (20, 135), (20, 137), (20, 138), (20, 165), (20, 167), (21, 15), (21, 17), (21, 18), (21, 45), (21, 47), (21, 48), 
(21, 75), (21, 77), (21, 78), (21, 105), (21, 107), (21, 108), (21, 135), (21, 137), (21, 138), (21, 165), (21, 167), (22, 15), (22, 17), (22, 18), (22, 45), (22, 47), (22, 48), (22, 75), (22, 77), (22, 78), (22, 105), (22, 107), (22, 108), (22, 135), (22, 137), (22, 138), (22, 165), (22, 167), (23, 15), (23, 17), 
(23, 18), (23, 45), (23, 47), (23, 48), (23, 75), (23, 77), (23, 78), (23, 105), (23, 107), (23, 108), (23, 135), (23, 137), (23, 138), (23, 165), (23, 167), (24, 15), (24, 17), (24, 18), (24, 45), (24, 47), (24, 48), (24, 75), (24, 77), (24, 78), (24, 105), (24, 107), (24, 108), (24, 135), (24, 137), (24, 138), 
(24, 165), (24, 167), (25, 15), (25, 17), (25, 18), (25, 45), (25, 47), (25, 48), (25, 75), (25, 77), (25, 78), (25, 105), (25, 107), (25, 108), (25, 135), (25, 137), (25, 138), (25, 165), (25, 167), (26, 15), (26, 17), (26, 18), (26, 45), (26, 47), (26, 48), (26, 75), (26, 77), (26, 78), (26, 105), (26, 107), 
(26, 108), (26, 135), (26, 137), (26, 138), (26, 165), (26, 167), (27, 15), (27, 17), (27, 18), (27, 45), (27, 47), (27, 48), (27, 75), (27, 77), (27, 78), (27, 105), (27, 107), (27, 108), (27, 135), (27, 137), (27, 138), (27, 165), (27, 167), (28, 15), (28, 17), (28, 18), (28, 45), (28, 47), (28, 48), (28, 75), 
(28, 77), (28, 78), (28, 80), (28, 82), (28, 83), (28, 85), (28, 87), (28, 88), (28, 90), (28, 92), (28, 93), (28, 95), (28, 97), (28, 98), (28, 100), (28, 102), (28, 103), (28, 105), (28, 107), (28, 108), (28, 135), (28, 137), (28, 138), (28, 165), (28, 167), (29, 15), (29, 17), (29, 18), (29, 45), (29, 47), 
(29, 48), (29, 75), (29, 77), (29, 78), (29, 80), (29, 82), (29, 83), (29, 85), (29, 87), (29, 88), (29, 90), (29, 92), (29, 93), (29, 95), (29, 97), (29, 98), (29, 100), (29, 102), (29, 103), (29, 105), (29, 107), (29, 108), (29, 135), (29, 137), (29, 138), (29, 165), (29, 167), (30, 15), (30, 17), (30, 18), 
(30, 45), (30, 47), (30, 48), (30, 75), (30, 77), (30, 78), (30, 80), (30, 82), (30, 83), (30, 85), (30, 87), (30, 88), (30, 90), (30, 92), (30, 93), (30, 95), (30, 97), (30, 98), (30, 100), (30, 102), (30, 103), (30, 105), (30, 107), (30, 108), (30, 135), (30, 137), (30, 138), (30, 165), (30, 167), (31, 15), 
(31, 17), (31, 18), (31, 45), (31, 47), (31, 48), (31, 75), (31, 77), (31, 78), (31, 80), (31, 82), (31, 83), (31, 85), (31, 87), (31, 88), (31, 90), (31, 92), (31, 93), (31, 95), (31, 97), (31, 98), (31, 100), (31, 102), (31, 103), (31, 105), (31, 107), (31, 108), (31, 135), (31, 137), (31, 138), (31, 165), 
(31, 167), (32, 15), (32, 17), (32, 18), (32, 45), (32, 47), (32, 48), (32, 50), (32, 52), (32, 53), (32, 55), (32, 57), (32, 58), (32, 60), (32, 62), (32, 63), (32, 65), (32, 67), (32, 68), (32, 70), (32, 72), (32, 73), (32, 75), (32, 77), (32, 78), (32, 105), (32, 107), (32, 108), (32, 135), (32, 137), 
(32, 138), (32, 165), (32, 167), (33, 15), (33, 17), (33, 18), (33, 45), (33, 47), (33, 48), (33, 50), (33, 52), (33, 53), (33, 55), (33, 57), (33, 58), (33, 60), (33, 62), (33, 63), (33, 65), (33, 67), (33, 68), (33, 70), (33, 72), (33, 73), (33, 75), (33, 77), (33, 78), (33, 105), (33, 107), (33, 108), 
(33, 135), (33, 137), (33, 138), (33, 165), (33, 167), (34, 15), (34, 17), (34, 18), (34, 45), (34, 47), (34, 48), (34, 50), (34, 52), (34, 53), (34, 55), (34, 57), (34, 58), (34, 60), (34, 62), (34, 63), (34, 65), (34, 67), (34, 68), (34, 70), (34, 72), (34, 73), (34, 75), (34, 77), (34, 78), (34, 105), 
(34, 107), (34, 108), (34, 135), (34, 137), (34, 138), (34, 165), (34, 167), (35, 15), (35, 17), (35, 18), (35, 45), (35, 47), (35, 48), (35, 50), (35, 52), (35, 53), (35, 55), (35, 57), (35, 58), (35, 60), (35, 62), (35, 63), (35, 65), (35, 67), (35, 68), (35, 70), (35, 72), (35, 73), (35, 75), (35, 77), 
(35, 78), (35, 105), (35, 107), (35, 108), (35, 135), (35, 137), (35, 138), (35, 165), (35, 167), (36, 15), (36, 17), (36, 18), (36, 45), (36, 47), (36, 48), (36, 75), (36, 77), (36, 78), (36, 105), (36, 107), (36, 108), (36, 110), (36, 112), (36, 113), (36, 115), (36, 117), (36, 118), (36, 120), (36, 122), (36, 123), 
(36, 125), (36, 127), (36, 128), (36, 130), (36, 132), (36, 133), (36, 135), (36, 137), (36, 138), (36, 165), (36, 167), (37, 15), (37, 17), (37, 18), (37, 45), (37, 47), (37, 48), (37, 75), (37, 77), (37, 78), (37, 105), (37, 107), (37, 108), (37, 110), (37, 112), (37, 113), (37, 115), (37, 117), (37, 118), (37, 120), 
(37, 122), (37, 123), (37, 125), (37, 127), (37, 128), (37, 130), (37, 132), (37, 133), (37, 135), (37, 137), (37, 138), (37, 165), (37, 167), (38, 15), (38, 17), (38, 18), (38, 45), (38, 47), (38, 48), (38, 75), (38, 77), (38, 78), (38, 105), (38, 107), (38, 108), (38, 110), (38, 112), (38, 113), (38, 115), (38, 117), 
(38, 118), (38, 120), (38, 122), (38, 123), (38, 125), (38, 127), (38, 128), (38, 130), (38, 132), (38, 133), (38, 135), (38, 137), (38, 138), (38, 165), (38, 167), (39, 15), (39, 17), (39, 18), (39, 45), (39, 47), (39, 48), (39, 75), (39, 77), (39, 78), (39, 105), (39, 107), (39, 108), (39, 110), (39, 112), (39, 113), 
(39, 115), (39, 117), (39, 118), (39, 120), (39, 122), (39, 123), (39, 125), (39, 127), (39, 128), (39, 130), (39, 132), (39, 133), (39, 135), (39, 137), (39, 138), (39, 165), (39, 167), (40, 15), (40, 17), (40, 18), (40, 20), (40, 22), (40, 23), (40, 25), (40, 27), (40, 28), (40, 30), (40, 32), (40, 33), (40, 35), 
(40, 37), (40, 38), (40, 40), (40, 42), (40, 43), (40, 45), (40, 47), (40, 48), (40, 75), (40, 77), (40, 78), (40, 105), (40, 107), (40, 108), (40, 135), (40, 137), (40, 138), (40, 140), (40, 142), (40, 143), (40, 145), (40, 147), (40, 148), (40, 150), (40, 152), (40, 153), (40, 155), (40, 157), (40, 158), (40, 160), 
(40, 162), (40, 163), (40, 165), (40, 167), (41, 15), (41, 17), (41, 18), (41, 20), (41, 22), (41, 23), (41, 25), (41, 27), (41, 28), (41, 30), (41, 32), (41, 33), (41, 35), (41, 37), (41, 38), (41, 40), (41, 42), (41, 43), (41, 45), (41, 47), (41, 48), (41, 75), (41, 77), (41, 78), (41, 105), (41, 107), 
(41, 108), (41, 135), (41, 137), (41, 138), (41, 140), (41, 142), (41, 143), (41, 145), (41, 147), (41, 148), (41, 150), (41, 152), (41, 153), (41, 155), (41, 157), (41, 158), (41, 160), (41, 162), (41, 163), (41, 165), (41, 167), (42, 15), (42, 17), (42, 18), (42, 20), (42, 22), (42, 23), (42, 25), (42, 27), (42, 28), 
(42, 30), (42, 32), (42, 33), (42, 35), (42, 37), (42, 38), (42, 40), (42, 42), (42, 43), (42, 45), (42, 47), (42, 48), (42, 75), (42, 77), (42, 78), (42, 105), (42, 107), (42, 108), (42, 135), (42, 137), (42, 138), (42, 140), (42, 142), (42, 143), (42, 145), (42, 147), (42, 148), (42, 150), (42, 152), (42, 153), 
(42, 155), (42, 157), (42, 158), (42, 160), (42, 162), (42, 163), (42, 165), (42, 167), (43, 15), (43, 17), (43, 18), (43, 20), (43, 22), (43, 23), (43, 25), (43, 27), (43, 28), (43, 30), (43, 32), (43, 33), (43, 35), (43, 37), (43, 38), (43, 40), (43, 42), (43, 43), (43, 45), (43, 47), (43, 48), (43, 75), 
(43, 77), (43, 78), (43, 105), (43, 107), (43, 108), (43, 135), (43, 137), (43, 138), (43, 140), (43, 142), (43, 143), (43, 145), (43, 147), (43, 148), (43, 150), (43, 152), (43, 153), (43, 155), (43, 157), (43, 158), (43, 160), (43, 162), (43, 163), (43, 165), (43, 167), (44, 15), (44, 17), (44, 18), (44, 45), (44, 47), 
(44, 48), (44, 75), (44, 77), (44, 78), (44, 105), (44, 107), (44, 108), (44, 135), (44, 137), (44, 138), (44, 165), (44, 167), (45, 15), (45, 17), (45, 18), (45, 45), (45, 47), (45, 48), (45, 75), (45, 77), (45, 78), (45, 105), (45, 107), (45, 108), (45, 135), (45, 137), (45, 138), (45, 165), (45, 167), (46, 15), 
(46, 17), (46, 18), (46, 45), (46, 47), (46, 48), (46, 75), (46, 77), (46, 78), (46, 105), (46, 107), (46, 108), (46, 135), (46, 137), (46, 138), (46, 165), (46, 167), (47, 15), (47, 17), (47, 18), (47, 45), (47, 47), (47, 48), (47, 75), (47, 77), (47, 78), (47, 105), (47, 107), (47, 108), (47, 135), (47, 137), 
(47, 138), (47, 165), (47, 167), (48, 15), (48, 17), (48, 18), (48, 45), (48, 47), (48, 48), (48, 75), (48, 77), (48, 78), (48, 105), (48, 107), (48, 108), (48, 135), (48, 137), (48, 138), (48, 165), (48, 167), (49, 15), (49, 17), (49, 18), (49, 45), (49, 47), (49, 48), (49, 75), (49, 77), (49, 78), (49, 105), 
(49, 107), (49, 108), (49, 135), (49, 137), (49, 138), (49, 165), (49, 167), (50, 15), (50, 17), (50, 18), (50, 45), (50, 47), (50, 48), (50, 75), (50, 77), (50, 78), (50, 105), (50, 107), (50, 108), (50, 135), (50, 137), (50, 138), (50, 165), (50, 167), (51, 15), (51, 17), (51, 18), (51, 45), (51, 47), (51, 48), 
(51, 75), (51, 77), (51, 78), (51, 105), (51, 107), (51, 108), (51, 135), (51, 137), (51, 138), (51, 165), (51, 167), (52, 15), (52, 17), (52, 18), (52, 45), (52, 47), (52, 48), (52, 50), (52, 52), (52, 53), (52, 55), (52, 57), (52, 58), (52, 60), (52, 62), (52, 63), (52, 65), (52, 67), (52, 68), (52, 70), 
(52, 72), (52, 73), (52, 75), (52, 77), (52, 78), (52, 105), (52, 107), (52, 108), (52, 135), (52, 137), (52, 138), (52, 165), (52, 167), (53, 15), (53, 17), (53, 18), (53, 45), (53, 47), (53, 48), (53, 50), (53, 52), (53, 53), (53, 55), (53, 57), (53, 58), (53, 60), (53, 62), (53, 63), (53, 65), (53, 67), 
(53, 68), (53, 70), (53, 72), (53, 73), (53, 75), (53, 77), (53, 78), (53, 105), (53, 107), (53, 108), (53, 135), (53, 137), (53, 138), (53, 165), (53, 167), (54, 15), (54, 17), (54, 18), (54, 45), (54, 47), (54, 48), (54, 50), (54, 52), (54, 53), (54, 55), (54, 57), (54, 58), (54, 60), (54, 62), (54, 63), 
(54, 65), (54, 67), (54, 68), (54, 70), (54, 72), (54, 73), (54, 75), (54, 77), (54, 78), (54, 105), (54, 107), (54, 108), (54, 135), (54, 137), (54, 138), (54, 165), (54, 167), (55, 15), (55, 17), (55, 18), (55, 45), (55, 47), (55, 48), (55, 50), (55, 52), (55, 53), (55, 55), (55, 57), (55, 58), (55, 60), 
(55, 62), (55, 63), (55, 65), (55, 67), (55, 68), (55, 70), (55, 72), (55, 73), (55, 75), (55, 77), (55, 78), (55, 105), (55, 107), (55, 108), (55, 135), (55, 137), (55, 138), (55, 165), (55, 167), (56, 15), (56, 17), (56, 18), (56, 20), (56, 22), (56, 23), (56, 25), (56, 27), (56, 28), (56, 30), (56, 32), 
(56, 33), (56, 35), (56, 37), (56, 38), (56, 40), (56, 42), (56, 43), (56, 45), (56, 47), (56, 48), (56, 75), (56, 77), (56, 78), (56, 105), (56, 107), (56, 108), (56, 135), (56, 137), (56, 138), (56, 165), (56, 167), (57, 15), (57, 17), (57, 18), (57, 20), (57, 22), (57, 23), (57, 25), (57, 27), (57, 28), 
(57, 30), (57, 32), (57, 33), (57, 35), (57, 37), (57, 38), (57, 40), (57, 42), (57, 43), (57, 45), (57, 47), (57, 48), (57, 75), (57, 77), (57, 78), (57, 105), (57, 107), (57, 108), (57, 135), (57, 137), (57, 138), (57, 165), (57, 167), (58, 15), (58, 17), (58, 18), (58, 20), (58, 22), (58, 23), (58, 25), 
(58, 27), (58, 28), (58, 30), (58, 32), (58, 33), (58, 35), (58, 37), (58, 38), (58, 40), (58, 42), (58, 43), (58, 45), (58, 47), (58, 48), (58, 75), (58, 77), (58, 78), (58, 105), (58, 107), (58, 108), (58, 135), (58, 137), (58, 138), (58, 165), (58, 167), (59, 15), (59, 17), (59, 18), (59, 20), (59, 22), 
(59, 23), (59, 25), (59, 27), (59, 28), (59, 30), (59, 32), (59, 33), (59, 35), (59, 37), (59, 38), (59, 40), (59, 42), (59, 43), (59, 45), (59, 47), (59, 48), (59, 75), (59, 77), (59, 78), (59, 105), (59, 107), (59, 108), (59, 135), (59, 137), (59, 138), (59, 165), (59, 167), (60, 15), (60, 17), (60, 18), 
(60, 45), (60, 47), (60, 48), (60, 75), (60, 77), (60, 78), (60, 80), (60, 82), (60, 83), (60, 85), (60, 87), (60, 88), (60, 90), (60, 92), (60, 93), (60, 95), (60, 97), (60, 98), (60, 100), (60, 102), (60, 103), (60, 105), (60, 107), (60, 108), (60, 135), (60, 137), (60, 138), (60, 165), (60, 167), (61, 15), 
(61, 17), (61, 18), (61, 45), (61, 47), (61, 48), (61, 75), (61, 77), (61, 78), (61, 80), (61, 82), (61, 83), (61, 85), (61, 87), (61, 88), (61, 90), (61, 92), (61, 93), (61, 95), (61, 97), (61, 98), (61, 100), (61, 102), (61, 103), (61, 105), (61, 107), (61, 108), (61, 135), (61, 137), (61, 138), (61, 165), 
(61, 167), (62, 15), (62, 17), (62, 18), (62, 45), (62, 47), (62, 48), (62, 75), (62, 77), (62, 78), (62, 80), (62, 82), (62, 83), (62, 85), (62, 87), (62, 88), (62, 90), (62, 92), (62, 93), (62, 95), (62, 97), (62, 98), (62, 100), (62, 102), (62, 103), (62, 105), (62, 107), (62, 108), (62, 135), (62, 137), 
(62, 138), (62, 165), (62, 167), (63, 15), (63, 17), (63, 18), (63, 45), (63, 47), (63, 48), (63, 75), (63, 77), (63, 78), (63, 80), (63, 82), (63, 83), (63, 85), (63, 87), (63, 88), (63, 90), (63, 92), (63, 93), (63, 95), (63, 97), (63, 98), (63, 100), (63, 102), (63, 103), (63, 105), (63, 107), (63, 108), 
(63, 135), (63, 137), (63, 138), (63, 165), (63, 167), (64, 15), (64, 17), (64, 18), (64, 45), (64, 47), (64, 48), (64, 50), (64, 52), (64, 53), (64, 55), (64, 57), (64, 58), (64, 60), (64, 62), (64, 63), (64, 65), (64, 67), (64, 68), (64, 70), (64, 72), (64, 73), (64, 75), (64, 77), (64, 78), (64, 105), 
(64, 107), (64, 108), (64, 135), (64, 137), (64, 138), (64, 140), (64, 142), (64, 143), (64, 145), (64, 147), (64, 148), (64, 150), (64, 152), (64, 153), (64, 155), (64, 157), (64, 158), (64, 160), (64, 162), (64, 163), (64, 165), (64, 167), (65, 15), (65, 17), (65, 18), (65, 45), (65, 47), (65, 48), (65, 50), (65, 52), 
(65, 53), (65, 55), (65, 57), (65, 58), (65, 60), (65, 62), (65, 63), (65, 65), (65, 67), (65, 68), (65, 70), (65, 72), (65, 73), (65, 75), (65, 77), (65, 78), (65, 105), (65, 107), (65, 108), (65, 135), (65, 137), (65, 138), (65, 140), (65, 142), (65, 143), (65, 145), (65, 147), (65, 148), (65, 150), (65, 152), 
(65, 153), (65, 155), (65, 157), (65, 158), (65, 160), (65, 162), (65, 163), (65, 165), (65, 167), (66, 15), (66, 17), (66, 18), (66, 45), (66, 47), (66, 48), (66, 50), (66, 52), (66, 53), (66, 55), (66, 57), (66, 58), (66, 60), (66, 62), (66, 63), (66, 65), (66, 67), (66, 68), (66, 70), (66, 72), (66, 73), 
(66, 75), (66, 77), (66, 78), (66, 105), (66, 107), (66, 108), (66, 135), (66, 137), (66, 138), (66, 140), (66, 142), (66, 143), (66, 145), (66, 147), (66, 148), (66, 150), (66, 152), (66, 153), (66, 155), (66, 157), (66, 158), (66, 160), (66, 162), (66, 163), (66, 165), (66, 167), (67, 15), (67, 17), (67, 18), (67, 45), 
(67, 47), (67, 48), (67, 50), (67, 52), (67, 53), (67, 55), (67, 57), (67, 58), (67, 60), (67, 62), (67, 63), (67, 65), (67, 67), (67, 68), (67, 70), (67, 72), (67, 73), (67, 75), (67, 77), (67, 78), (67, 105), (67, 107), (67, 108), (67, 135), (67, 137), (67, 138), (67, 140), (67, 142), (67, 143), (67, 145), 
(67, 147), (67, 148), (67, 150), (67, 152), (67, 153), (67, 155), (67, 157), (67, 158), (67, 160), (67, 162), (67, 163), (67, 165), (67, 167), (68, 15), (68, 17), (68, 18), (68, 45), (68, 47), (68, 48), (68, 75), (68, 77), (68, 78), (68, 105), (68, 107), (68, 108), (68, 135), (68, 137), (68, 138), (68, 165), (68, 167), 
(69, 15), (69, 17), (69, 18), (69, 45), (69, 47), (69, 48), (69, 75), (69, 77), (69, 78), (69, 105), (69, 107), (69, 108), (69, 135), (69, 137), (69, 138), (69, 165), (69, 167), (70, 15), (70, 17), (70, 18), (70, 45), (70, 47), (70, 48), (70, 75), (70, 77), (70, 78), (70, 105), (70, 107), (70, 108), (70, 135), 
(70, 137), (70, 138), (70, 165), (70, 167), (71, 15), (71, 17), (71, 18), (71, 45), (71, 47), (71, 48), (71, 75), (71, 77), (71, 78), (71, 105), (71, 107), (71, 108), (71, 135), (71, 137), (71, 138), (71, 165), (71, 167), (72, 15), (72, 17), (72, 18), (72, 20), (72, 22), (72, 23), (72, 25), (72, 27), (72, 28), 
(72, 30), (72, 32), (72, 33), (72, 35), (72, 37), (72, 38), (72, 40), (72, 42), (72, 43), (72, 45), (72, 47), (72, 48), (72, 75), (72, 77), (72, 78), (72, 105), (72, 107), (72, 108), (72, 110), (72, 112), (72, 113), (72, 115), (72, 117), (72, 118), (72, 120), (72, 122), (72, 123), (72, 125), (72, 127), (72, 128), 
(72, 130), (72, 132), (72, 133), (72, 135), (72, 137), (72, 138), (72, 165), (72, 167), (73, 15), (73, 17), (73, 18), (73, 20), (73, 22), (73, 23), (73, 25), (73, 27), (73, 28), (73, 30), (73, 32), (73, 33), (73, 35), (73, 37), (73, 38), (73, 40), (73, 42), (73, 43), (73, 45), (73, 47), (73, 48), (73, 75), 
(73, 77), (73, 78), (73, 105), (73, 107), (73, 108), (73, 110), (73, 112), (73, 113), (73, 115), (73, 117), (73, 118), (73, 120), (73, 122), (73, 123), (73, 125), (73, 127), (73, 128), (73, 130), (73, 132), (73, 133), (73, 135), (73, 137), (73, 138), (73, 165), (73, 167), (74, 15), (74, 17), (74, 18), (74, 20), (74, 22), 
(74, 23), (74, 25), (74, 27), (74, 28), (74, 30), (74, 32), (74, 33), (74, 35), (74, 37), (74, 38), (74, 40), (74, 42), (74, 43), (74, 45), (74, 47), (74, 48), (74, 75), (74, 77), (74, 78), (74, 105), (74, 107), (74, 108), (74, 110), (74, 112), (74, 113), (74, 115), (74, 117), (74, 118), (74, 120), (74, 122), 
(74, 123), (74, 125), (74, 127), (74, 128), (74, 130), (74, 132), (74, 133), (74, 135), (74, 137), (74, 138), (74, 165), (74, 167), (75, 15), (75, 17), (75, 18), (75, 20), (75, 22), (75, 23), (75, 25), (75, 27), (75, 28), (75, 30), (75, 32), (75, 33), (75, 35), (75, 37), (75, 38), (75, 40), (75, 42), (75, 43), 
(75, 45), (75, 47), (75, 48), (75, 75), (75, 77), (75, 78), (75, 105), (75, 107), (75, 108), (75, 110), (75, 112), (75, 113), (75, 115), (75, 117), (75, 118), (75, 120), (75, 122), (75, 123), (75, 125), (75, 127), (75, 128), (75, 130), (75, 132), (75, 133), (75, 135), (75, 137), (75, 138), (75, 165), (75, 167), (76, 15), 
(76, 17), (76, 18), (76, 45), (76, 47), (76, 48), (76, 75), (76, 77), (76, 78), (76, 105), (76, 107), (76, 108), (76, 135), (76, 137), (76, 138), (76, 165), (76, 167), (77, 15), (77, 17), (77, 18), (77, 45), (77, 47), (77, 48), (77, 75), (77, 77), (77, 78), (77, 105), (77, 107), (77, 108), (77, 135), (77, 137), 
(77, 138), (77, 165), (77, 167), (78, 15), (78, 17), (78, 18), (78, 45), (78, 47), (78, 48), (78, 75), (78, 77), (78, 78), (78, 105), (78, 107), (78, 108), (78, 135), (78, 137), (78, 138), (78, 165), (78, 167), (79, 15), (79, 17), (79, 18), (79, 45), (79, 47), (79, 48), (79, 75), (79, 77), (79, 78), (79, 105), 
(79, 107), (79, 108), (79, 135), (79, 137), (79, 138), (79, 165), (79, 167), (80, 15), (80, 17), (80, 18), (80, 45), (80, 47), (80, 48), (80, 75), (80, 77), (80, 78), (80, 105), (80, 107), (80, 108), (80, 135), (80, 137), (80, 138), (80, 165), (80, 167), (81, 15), (81, 17), (81, 18), (81, 45), (81, 47), (81, 48), 
(81, 75), (81, 77), (81, 78), (81, 105), (81, 107), (81, 108), (81, 135), (81, 137), (81, 138), (81, 165), (81, 167), (82, 15), (82, 17), (82, 18), (82, 45), (82, 47), (82, 48), (82, 75), (82, 77), (82, 78), (82, 105), (82, 107), (82, 108), (82, 135), (82, 137), (82, 138), (82, 165), (82, 167), (83, 15), (83, 17), 
(83, 18), (83, 45), (83, 47), (83, 48), (83, 75), (83, 77), (83, 78), (83, 105), (83, 107), (83, 108), (83, 135), (83, 137), (83, 138), (83, 165), (83, 167), (84, 15), (84, 17), (84, 18), (84, 20), (84, 22), (84, 23), (84, 25), (84, 27), (84, 28), (84, 30), (84, 32), (84, 33), (84, 35), (84, 37), (84, 38), 
(84, 40), (84, 42), (84, 43), (84, 45), (84, 47), (84, 48), (84, 75), (84, 77), (84, 78), (84, 105), (84, 107), (84, 108), (84, 110), (84, 112), (84, 113), (84, 115), (84, 117), (84, 118), (84, 120), (84, 122), (84, 123), (84, 125), (84, 127), (84, 128), (84, 130), (84, 132), (84, 133), (84, 135), (84, 137), (84, 138), 
(84, 165), (84, 167), (85, 15), (85, 17), (85, 18), (85, 20), (85, 22), (85, 23), (85, 25), (85, 27), (85, 28), (85, 30), (85, 32), (85, 33), (85, 35), (85, 37), (85, 38), (85, 40), (85, 42), (85, 43), (85, 45), (85, 47), (85, 48), (85, 75), (85, 77), (85, 78), (85, 105), (85, 107), (85, 108), (85, 110), 
(85, 112), (85, 113), (85, 115), (85, 117), (85, 118), (85, 120), (85, 122), (85, 123), (85, 125), (85, 127), (85, 128), (85, 130), (85, 132), (85, 133), (85, 135), (85, 137), (85, 138), (85, 165), (85, 167), (86, 15), (86, 17), (86, 18), (86, 20), (86, 22), (86, 23), (86, 25), (86, 27), (86, 28), (86, 30), (86, 32), 
(86, 33), (86, 35), (86, 37), (86, 38), (86, 40), (86, 42), (86, 43), (86, 45), (86, 47), (86, 48), (86, 75), (86, 77), (86, 78), (86, 105), (86, 107), (86, 108), (86, 110), (86, 112), (86, 113), (86, 115), (86, 117), (86, 118), (86, 120), (86, 122), (86, 123), (86, 125), (86, 127), (86, 128), (86, 130), (86, 132), 
(86, 133), (86, 135), (86, 137), (86, 138), (86, 165), (86, 167), (87, 15), (87, 17), (87, 18), (87, 20), (87, 22), (87, 23), (87, 25), (87, 27), (87, 28), (87, 30), (87, 32), (87, 33), (87, 35), (87, 37), (87, 38), (87, 40), (87, 42), (87, 43), (87, 45), (87, 47), (87, 48), (87, 75), (87, 77), (87, 78), 
(87, 105), (87, 107), (87, 108), (87, 110), (87, 112), (87, 113), (87, 115), (87, 117), (87, 118), (87, 120), (87, 122), (87, 123), (87, 125), (87, 127), (87, 128), (87, 130), (87, 132), (87, 133), (87, 135), (87, 137), (87, 138), (87, 165), (87, 167), (88, 15), (88, 17), (88, 18), (88, 45), (88, 47), (88, 48), (88, 75), 
(88, 77), (88, 78), (88, 105), (88, 107), (88, 108), (88, 135), (88, 137), (88, 138), (88, 165), (88, 167), (89, 15), (89, 17), (89, 18), (89, 45), (89, 47), (89, 48), (89, 75), (89, 77), (89, 78), (89, 105), (89, 107), (89, 108), (89, 135), (89, 137), (89, 138), (89, 165), (89, 167), (90, 15), (90, 17), (90, 18), 
(90, 45), (90, 47), (90, 48), (90, 75), (90, 77), (90, 78), (90, 105), (90, 107), (90, 108), (90, 135), (90, 137), (90, 138), (90, 165), (90, 167), (91, 15), (91, 17), (91, 18), (91, 45), (91, 47), (91, 48), (91, 75), (91, 77), (91, 78), (91, 105), (91, 107), (91, 108), (91, 135), (91, 137), (91, 138), (91, 165), 
(91, 167), (92, 15), (92, 17), (92, 18), (92, 45), (92, 47), (92, 48), (92, 50), (92, 52), (92, 53), (92, 55), (92, 57), (92, 58), (92, 60), (92, 62), (92, 63), (92, 65), (92, 67), (92, 68), (92, 70), (92, 72), (92, 73), (92, 75), (92, 77), (92, 78), (92, 105), (92, 107), (92, 108), (92, 135), (92, 137), 
(92, 138), (92, 140), (92, 142), (92, 143), (92, 145), (92, 147), (92, 148), (92, 150), (92, 152), (92, 153), (92, 155), (92, 157), (92, 158), (92, 160), (92, 162), (92, 163), (92, 165), (92, 167), (93, 15), (93, 17), (93, 18), (93, 45), (93, 47), (93, 48), (93, 50), (93, 52), (93, 53), (93, 55), (93, 57), (93, 58), 
(93, 60), (93, 62), (93, 63), (93, 65), (93, 67), (93, 68), (93, 70), (93, 72), (93, 73), (93, 75), (93, 77), (93, 78), (93, 105), (93, 107), (93, 108), (93, 135), (93, 137), (93, 138), (93, 140), (93, 142), (93, 143), (93, 145), (93, 147), (93, 148), (93, 150), (93, 152), (93, 153), (93, 155), (93, 157), (93, 158), 
(93, 160), (93, 162), (93, 163), (93, 165), (93, 167), (94, 15), (94, 17), (94, 18), (94, 45), (94, 47), (94, 48), (94, 50), (94, 52), (94, 53), (94, 55), (94, 57), (94, 58), (94, 60), (94, 62), (94, 63), (94, 65), (94, 67), (94, 68), (94, 70), (94, 72), (94, 73), (94, 75), (94, 77), (94, 78), (94, 105), 
(94, 107), (94, 108), (94, 135), (94, 137), (94, 138), (94, 140), (94, 142), (94, 143), (94, 145), (94, 147), (94, 148), (94, 150), (94, 152), (94, 153), (94, 155), (94, 157), (94, 158), (94, 160), (94, 162), (94, 163), (94, 165), (94, 167), (95, 15), (95, 17), (95, 18), (95, 45), (95, 47), (95, 48), (95, 50), (95, 52), 
(95, 53), (95, 55), (95, 57), (95, 58), (95, 60), (95, 62), (95, 63), (95, 65), (95, 67), (95, 68), (95, 70), (95, 72), (95, 73), (95, 75), (95, 77), (95, 78), (95, 105), (95, 107), (95, 108), (95, 135), (95, 137), (95, 138), (95, 140), (95, 142), (95, 143), (95, 145), (95, 147), (95, 148), (95, 150), (95, 152), 
(95, 153), (95, 155), (95, 157), (95, 158), (95, 160), (95, 162), (95, 163), (95, 165), (95, 167), (96, 15), (96, 17), (96, 18), (96, 45), (96, 47), (96, 48), (96, 75), (96, 77), (96, 78), (96, 80), (96, 82), (96, 83), (96, 85), (96, 87), (96, 88), (96, 90), (96, 92), (96, 93), (96, 95), (96, 97), (96, 98), 
(96, 100), (96, 102), (96, 103), (96, 105), (96, 107), (96, 108), (96, 135), (96, 137), (96, 138), (96, 165), (96, 167), (97, 15), (97, 17), (97, 18), (97, 45), (97, 47), (97, 48), (97, 75), (97, 77), (97, 78), (97, 80), (97, 82), (97, 83), (97, 85), (97, 87), (97, 88), (97, 90), (97, 92), (97, 93), (97, 95), 
(97, 97), (97, 98), (97, 100), (97, 102), (97, 103), (97, 105), (97, 107), (97, 108), (97, 135), (97, 137), (97, 138), (97, 165), (97, 167), (98, 15), (98, 17), (98, 18), (98, 45), (98, 47), (98, 48), (98, 75), (98, 77), (98, 78), (98, 80), (98, 82), (98, 83), (98, 85), (98, 87), (98, 88), (98, 90), (98, 92), 
(98, 93), (98, 95), (98, 97), (98, 98), (98, 100), (98, 102), (98, 103), (98, 105), (98, 107), (98, 108), (98, 135), (98, 137), (98, 138), (98, 165), (98, 167), (99, 15), (99, 17), (99, 18), (99, 45), (99, 47), (99, 48), (99, 75), (99, 77), (99, 78), (99, 80), (99, 82), (99, 83), (99, 85), (99, 87), (99, 88), 
(99, 90), (99, 92), (99, 93), (99, 95), (99, 97), (99, 98), (99, 100), (99, 102), (99, 103), (99, 105), (99, 107), (99, 108), (99, 135), (99, 137), (99, 138), (99, 165), (99, 167), (100, 15), (100, 17), (100, 18), (100, 20), (100, 22), (100, 23), (100, 25), (100, 27), (100, 28), (100, 30), (100, 32), (100, 33), (100, 35), 
(100, 37), (100, 38), (100, 40), (100, 42), (100, 43), (100, 45), (100, 47), (100, 48), (100, 75), (100, 77), (100, 78), (100, 105), (100, 107), (100, 108), (100, 135), (100, 137), (100, 138), (100, 165), (100, 167), (101, 15), (101, 17), (101, 18), (101, 20), (101, 22), (101, 23), (101, 25), (101, 27), (101, 28), (101, 30), (101, 32), 
(101, 33), (101, 35), (101, 37), (101, 38), (101, 40), (101, 42), (101, 43), (101, 45), (101, 47), (101, 48), (101, 75), (101, 77), (101, 78), (101, 105), (101, 107), (101, 108), (101, 135), (101, 137), (101, 138), (101, 165), (101, 167), (102, 15), (102, 17), (102, 18), (102, 20), (102, 22), (102, 23), (102, 25), (102, 27), (102, 28), 
(102, 30), (102, 32), (102, 33), (102, 35), (102, 37), (102, 38), (102, 40), (102, 42), (102, 43), (102, 45), (102, 47), (102, 48), (102, 75), (102, 77), (102, 78), (102, 105), (102, 107), (102, 108), (102, 135), (102, 137), (102, 138), (102, 165), (102, 167), (103, 15), (103, 17), (103, 18), (103, 20), (103, 22), (103, 23), (103, 25), 
(103, 27), (103, 28), (103, 30), (103, 32), (103, 33), (103, 35), (103, 37), (103, 38), (103, 40), (103, 42), (103, 43), (103, 45), (103, 47), (103, 48), (103, 75), (103, 77), (103, 78), (103, 105), (103, 107), (103, 108), (103, 135), (103, 137), (103, 138), (103, 165), (103, 167), (104, 15), (104, 17), (104, 18), (104, 45), (104, 47), 
(104, 48), (104, 50), (104, 52), (104, 53), (104, 55), (104, 57), (104, 58), (104, 60), (104, 62), (104, 63), (104, 65), (104, 67), (104, 68), (104, 70), (104, 72), (104, 73), (104, 75), (104, 77), (104, 78), (104, 105), (104, 107), (104, 108), (104, 135), (104, 137), (104, 138), (104, 165), (104, 167), (105, 15), (105, 17), (105, 18), 
(105, 45), (105, 47), (105, 48), (105, 50), (105, 52), (105, 53), (105, 55), (105, 57), (105, 58), (105, 60), (105, 62), (105, 63), (105, 65), (105, 67), (105, 68), (105, 70), (105, 72), (105, 73), (105, 75), (105, 77), (105, 78), (105, 105), (105, 107), (105, 108), (105, 135), (105, 137), (105, 138), (105, 165), (105, 167), (106, 15), 
(106, 17), (106, 18), (106, 45), (106, 47), (106, 48), (106, 50), (106, 52), (106, 53), (106, 55), (106, 57), (106, 58), (106, 60), (106, 62), (106, 63), (106, 65), (106, 67), (106, 68), (106, 70), (106, 72), (106, 73), (106, 75), (106, 77), (106, 78), (106, 105), (106, 107), (106, 108), (106, 135), (106, 137), (106, 138), (106, 165), 
(106, 167), (107, 15), (107, 17), (107, 18), (107, 45), (107, 47), (107, 48), (107, 50), (107, 52), (107, 53), (107, 55), (107, 57), (107, 58), (107, 60), (107, 62), (107, 63), (107, 65), (107, 67), (107, 68), (107, 70), (107, 72), (107, 73), (107, 75), (107, 77), (107, 78), (107, 105), (107, 107), (107, 108), (107, 135), (107, 137), 
(107, 138), (107, 165), (107, 167), (108, 15), (108, 17), (108, 18), (108, 45), (108, 47), (108, 48), (108, 75), (108, 77), (108, 78), (108, 105), (108, 107), (108, 108), (108, 135), (108, 137), (108, 138), (108, 165), (108, 167), (109, 15), (109, 17), (109, 18), (109, 45), (109, 47), (109, 48), (109, 75), (109, 77), (109, 78), (109, 105), 
(109, 107), (109, 108), (109, 135), (109, 137), (109, 138), (109, 165), (109, 167), (110, 15), (110, 17), (110, 18), (110, 45), (110, 47), (110, 48), (110, 75), (110, 77), (110, 78), (110, 105), (110, 107), (110, 108), (110, 135), (110, 137), (110, 138), (110, 165), (110, 167), (111, 15), (111, 17), (111, 18), (111, 45), (111, 47), (111, 48), 
(111, 75), (111, 77), (111, 78), (111, 105), (111, 107), (111, 108), (111, 135), (111, 137), (111, 138), (111, 165), (111, 167), (112, 15), (112, 17), (112, 18), (112, 45), (112, 47), (112, 48), (112, 75), (112, 77), (112, 78), (112, 105), (112, 107), (112, 108), (112, 135), (112, 137), (112, 138), (112, 165), (112, 167), (113, 15), (113, 17), 
(113, 18), (113, 45), (113, 47), (113, 48), (113, 75), (113, 77), (113, 78), (113, 105), (113, 107), (113, 108), (113, 135), (113, 137), (113, 138), (113, 165), (113, 167), (114, 15), (114, 17), (114, 18), (114, 45), (114, 47), (114, 48), (114, 75), (114, 77), (114, 78), (114, 105), (114, 107), (114, 108), (114, 135), (114, 137), (114, 138), 
(114, 165), (114, 167), (115, 15), (115, 17), (115, 18), (115, 45), (115, 47), (115, 48), (115, 75), (115, 77), (115, 78), (115, 105), (115, 107), (115, 108), (115, 135), (115, 137), (115, 138), (115, 165), (115, 167), (116, 15), (116, 17), (116, 18), (116, 20), (116, 22), (116, 23), (116, 25), (116, 27), (116, 28), (116, 30), (116, 32), 
(116, 33), (116, 35), (116, 37), (116, 38), (116, 40), (116, 42), (116, 43), (116, 45), (116, 47), (116, 48), (116, 75), (116, 77), (116, 78), (116, 105), (116, 107), (116, 108), (116, 135), (116, 137), (116, 138), (116, 140), (116, 142), (116, 143), (116, 145), (116, 147), (116, 148), (116, 150), (116, 152), (116, 153), (116, 155), (116, 157), 
(116, 158), (116, 160), (116, 162), (116, 163), (116, 165), (116, 167), (117, 15), (117, 17), (117, 18), (117, 20), (117, 22), (117, 23), (117, 25), (117, 27), (117, 28), (117, 30), (117, 32), (117, 33), (117, 35), (117, 37), (117, 38), (117, 40), (117, 42), (117, 43), (117, 45), (117, 47), (117, 48), (117, 75), (117, 77), (117, 78), 
(117, 105), (117, 107), (117, 108), (117, 135), (117, 137), (117, 138), (117, 140), (117, 142), (117, 143), (117, 145), (117, 147), (117, 148), (117, 150), (117, 152), (117, 153), (117, 155), (117, 157), (117, 158), (117, 160), (117, 162), (117, 163), (117, 165), (117, 167), (118, 15), (118, 17), (118, 18), (118, 20), (118, 22), (118, 23), (118, 25), 
(118, 27), (118, 28), (118, 30), (118, 32), (118, 33), (118, 35), (118, 37), (118, 38), (118, 40), (118, 42), (118, 43), (118, 45), (118, 47), (118, 48), (118, 75), (118, 77), (118, 78), (118, 105), (118, 107), (118, 108), (118, 135), (118, 137), (118, 138), (118, 140), (118, 142), (118, 143), (118, 145), (118, 147), (118, 148), (118, 150), 
(118, 152), (118, 153), (118, 155), (118, 157), (118, 158), (118, 160), (118, 162), (118, 163), (118, 165), (118, 167), (119, 15), (119, 17), (119, 18), (119, 20), (119, 22), (119, 23), (119, 25), (119, 27), (119, 28), (119, 30), (119, 32), (119, 33), (119, 35), (119, 37), (119, 38), (119, 40), (119, 42), (119, 43), (119, 45), (119, 47), 
(119, 48), (119, 75), (119, 77), (119, 78), (119, 105), (119, 107), (119, 108), (119, 135), (119, 137), (119, 138), (119, 140), (119, 142), (119, 143), (119, 145), (119, 147), (119, 148), (119, 150), (119, 152), (119, 153), (119, 155), (119, 157), (119, 158), (119, 160), (119, 162), (119, 163), (119, 165), (119, 167), (120, 15), (120, 17), (120, 18), 
(120, 45), (120, 47), (120, 48), (120, 75), (120, 77), (120, 78), (120, 105), (120, 107), (120, 108), (120, 110), (120, 112), (120, 113), (120, 115), (120, 117), (120, 118), (120, 120), (120, 122), (120, 123), (120, 125), (120, 127), (120, 128), (120, 130), (120, 132), (120, 133), (120, 135), (120, 137), (120, 138), (120, 165), (120, 167), (121, 15), 
(121, 17), (121, 18), (121, 45), (121, 47), (121, 48), (121, 75), (121, 77), (121, 78), (121, 105), (121, 107), (121, 108), (121, 110), (121, 112), (121, 113), (121, 115), (121, 117), (121, 118), (121, 120), (121, 122), (121, 123), (121, 125), (121, 127), (121, 128), (121, 130), (121, 132), (121, 133), (121, 135), (121, 137), (121, 138), (121, 165), 
(121, 167), (122, 15), (122, 17), (122, 18), (122, 45), (122, 47), (122, 48), (122, 75), (122, 77), (122, 78), (122, 105), (122, 107), (122, 108), (122, 110), (122, 112), (122, 113), (122, 115), (122, 117), (122, 118), (122, 120), (122, 122), (122, 123), (122, 125), (122, 127), (122, 128), (122, 130), (122, 132), (122, 133), (122, 135), (122, 137), 
(122, 138), (122, 165), (122, 167), (123, 15), (123, 17), (123, 18), (123, 45), (123, 47), (123, 48), (123, 75), (123, 77), (123, 78), (123, 105), (123, 107), (123, 108), (123, 110), (123, 112), (123, 113), (123, 115), (123, 117), (123, 118), (123, 120), (123, 122), (123, 123), (123, 125), (123, 127), (123, 128), (123, 130), (123, 132), (123, 133), 
(123, 135), (123, 137), (123, 138), (123, 165), (123, 167), (124, 15), (124, 17), (124, 18), (124, 45), (124, 47), (124, 48), (124, 50), (124, 52), (124, 53), (124, 55), (124, 57), (124, 58), (124, 60), (124, 62), (124, 63), (124, 65), (124, 67), (124, 68), (124, 70), (124, 72), (124, 73), (124, 75), (124, 77), (124, 78), (124, 105), 
(124, 107), (124, 108), (124, 135), (124, 137), (124, 138), (124, 165), (124, 167), (125, 15), (125, 17), (125, 18), (125, 45), (125, 47), (125, 48), (125, 50), (125, 52), (125, 53), (125, 55), (125, 57), (125, 58), (125, 60), (125, 62), (125, 63), (125, 65), (125, 67), (125, 68), (125, 70), (125, 72), (125, 73), (125, 75), (125, 77), 
(125, 78), (125, 105), (125, 107), (125, 108), (125, 135), (125, 137), (125, 138), (125, 165), (125, 167), (126, 15), (126, 17), (126, 18), (126, 45), (126, 47), (126, 48), (126, 50), (126, 52), (126, 53), (126, 55), (126, 57), (126, 58), (126, 60), (126, 62), (126, 63), (126, 65), (126, 67), (126, 68), (126, 70), (126, 72), (126, 73), 
(126, 75), (126, 77), (126, 78), (126, 105), (126, 107), (126, 108), (126, 135), (126, 137), (126, 138), (126, 165), (126, 167), (127, 15), (127, 17), (127, 18), (127, 45), (127, 47), (127, 48), (127, 50), (127, 52), (127, 53), (127, 55), (127, 57), (127, 58), (127, 60), (127, 62), (127, 63), (127, 65), (127, 67), (127, 68), (127, 70), 
(127, 72), (127, 73), (127, 75), (127, 77), (127, 78), (127, 105), (127, 107), (127, 108), (127, 135), (127, 137), (127, 138), (127, 165), (127, 167), (128, 15), (128, 17), (128, 18), (128, 45), (128, 47), (128, 48), (128, 75), (128, 77), (128, 78), (128, 80), (128, 82), (128, 83), (128, 85), (128, 87), (128, 88), (128, 90), (128, 92), 
(128, 93), (128, 95), (128, 97), (128, 98), (128, 100), (128, 102), (128, 103), (128, 105), (128, 107), (128, 108), (128, 135), (128, 137), (128, 138), (128, 165), (128, 167), (129, 15), (129, 17), (129, 18), (129, 45), (129, 47), (129, 48), (129, 75), (129, 77), (129, 78), (129, 80), (129, 82), (129, 83), (129, 85), (129, 87), (129, 88), 
(129, 90), (129, 92), (129, 93), (129, 95), (129, 97), (129, 98), (129, 100), (129, 102), (129, 103), (129, 105), (129, 107), (129, 108), (129, 135), (129, 137), (129, 138), (129, 165), (129, 167), (130, 15), (130, 17), (130, 18), (130, 45), (130, 47), (130, 48), (130, 75), (130, 77), (130, 78), (130, 80), (130, 82), (130, 83), (130, 85), 
(130, 87), (130, 88), (130, 90), (130, 92), (130, 93), (130, 95), (130, 97), (130, 98), (130, 100), (130, 102), (130, 103), (130, 105), (130, 107), (130, 108), (130, 135), (130, 137), (130, 138), (130, 165), (130, 167), (131, 15), (131, 17), (131, 18), (131, 45), (131, 47), (131, 48), (131, 75), (131, 77), (131, 78), (131, 80), (131, 82), 
(131, 83), (131, 85), (131, 87), (131, 88), (131, 90), (131, 92), (131, 93), (131, 95), (131, 97), (131, 98), (131, 100), (131, 102), (131, 103), (131, 105), (131, 107), (131, 108), (131, 135), (131, 137), (131, 138), (131, 165), (131, 167), (132, 15), (132, 17), (132, 18), (132, 45), (132, 47), (132, 48), (132, 75), (132, 77), (132, 78), 
(132, 105), (132, 107), (132, 108), (132, 135), (132, 137), (132, 138), (132, 165), (132, 167), (133, 15), (133, 17), (133, 18), (133, 45), (133, 47), (133, 48), (133, 75), (133, 77), (133, 78), (133, 105), (133, 107), (133, 108), (133, 135), (133, 137), (133, 138), (133, 165), (133, 167), (134, 15), (134, 17), (134, 18), (134, 45), (134, 47), 
(134, 48), (134, 75), (134, 77), (134, 78), (134, 105), (134, 107), (134, 108), (134, 135), (134, 137), (134, 138), (134, 165), (134, 167), (135, 15), (135, 17), (135, 18), (135, 45), (135, 47), (135, 48), (135, 75), (135, 77), (135, 78), (135, 105), (135, 107), (135, 108), (135, 135), (135, 137), (135, 138), (135, 165), (135, 167), (136, 15), 
(136, 17), (136, 18), (136, 45), (136, 47), (136, 48), (136, 75), (136, 77), (136, 78), (136, 105), (136, 107), (136, 108), (136, 135), (136, 137), (136, 138), (136, 165), (136, 167), (137, 15), (137, 17), (137, 18), (137, 45), (137, 47), (137, 48), (137, 75), (137, 77), (137, 78), (137, 105), (137, 107), (137, 108), (137, 135), (137, 137), 
(137, 138), (137, 165), (137, 167), (138, 15), (138, 17), (138, 18), (138, 45), (138, 47), (138, 48), (138, 75), (138, 77), (138, 78), (138, 105), (138, 107), (138, 108), (138, 135), (138, 137), (138, 138), (138, 165), (138, 167), (139, 15), (139, 17), (139, 18), (139, 45), (139, 47), (139, 48), (139, 75), (139, 77), (139, 78), (139, 105), 
(139, 107), (139, 108), (139, 135), (139, 137), (139, 138), (139, 165), (139, 167), (140, 15), (140, 17), (140, 18), (140, 20), (140, 22), (140, 23), (140, 25), (140, 27), (140, 28), (140, 30), (140, 32), (140, 33), (140, 35), (140, 37), (140, 38), (140, 40), (140, 42), (140, 43), (140, 45), (140, 47), (140, 48), (140, 50), (140, 52), 
(140, 53), (140, 55), (140, 57), (140, 58), (140, 60), (140, 62), (140, 63), (140, 65), (140, 67), (140, 68), (140, 70), (140, 72), (140, 73), (140, 75), (140, 77), (140, 78), (140, 80), (140, 82), (140, 83), (140, 85), (140, 87), (140, 88), (140, 90), (140, 92), (140, 93), (140, 95), (140, 97), (140, 98), (140, 100), (140, 102), 
(140, 103), (140, 105), (140, 107), (140, 108), (140, 110), (140, 112), (140, 113), (140, 115), (140, 117), (140, 118), (140, 120), (140, 122), (140, 123), (140, 125), (140, 127), (140, 128), (140, 130), (140, 132), (140, 133), (140, 135), (140, 137), (140, 138), (140, 140), (140, 142), (140, 143), (140, 145), (140, 147), (140, 148), (140, 150), (140, 152), 
(140, 153), (140, 155), (140, 157), (140, 158), (140, 160), (140, 162), (140, 163), (140, 165), (140, 167), (141, 15), (141, 17), (141, 18), (141, 20), (141, 22), (141, 23), (141, 25), (141, 27), (141, 28), (141, 30), (141, 32), (141, 33), (141, 35), (141, 37), (141, 38), (141, 40), (141, 42), (141, 43), (141, 45), (141, 47), (141, 48), 
(141, 50), (141, 52), (141, 53), (141, 55), (141, 57), (141, 58), (141, 60), (141, 62), (141, 63), (141, 65), (141, 67), (141, 68), (141, 70), (141, 72), (141, 73), (141, 75), (141, 77), (141, 78), (141, 80), (141, 82), (141, 83), (141, 85), (141, 87), (141, 88), (141, 90), (141, 92), (141, 93), (141, 95), (141, 97), (141, 98), 
(141, 100), (141, 102), (141, 103), (141, 105), (141, 107), (141, 108), (141, 110), (141, 112), (141, 113), (141, 115), (141, 117), (141, 118), (141, 120), (141, 122), (141, 123), (141, 125), (141, 127), (141, 128), (141, 130), (141, 132), (141, 133), (141, 135), (141, 137), (141, 138), (141, 140), (141, 142), (141, 143), (141, 145), (141, 147), (141, 148), 
(141, 150), (141, 152), (141, 153), (141, 155), (141, 157), (141, 158), (141, 160), (141, 162), (141, 163), (141, 165), (141, 167), (142, 15), (142, 17), (142, 18), (142, 20), (142, 22), (142, 23), (142, 25), (142, 27), (142, 28), (142, 30), (142, 32), (142, 33), (142, 35), (142, 37), (142, 38), (142, 40), (142, 42), (142, 43), (142, 45), 
(142, 47), (142, 48), (142, 50), (142, 52), (142, 53), (142, 55), (142, 57), (142, 58), (142, 60), (142, 62), (142, 63), (142, 65), (142, 67), (142, 68), (142, 70), (142, 72), (142, 73), (142, 75), (142, 77), (142, 78), (142, 80), (142, 82), (142, 83), (142, 85), (142, 87), (142, 88), (142, 90), (142, 92), (142, 93), (142, 95), 
(142, 97), (142, 98), (142, 100), (142, 102), (142, 103), (142, 105), (142, 107), (142, 108), (142, 110), (142, 112), (142, 113), (142, 115), (142, 117), (142, 118), (142, 120), (142, 122), (142, 123), (142, 125), (142, 127), (142, 128), (142, 130), (142, 132), (142, 133), (142, 135), (142, 137), (142, 138), (142, 140), (142, 142), (142, 143), (142, 145), 
(142, 147), (142, 148), (142, 150), (142, 152), (142, 153), (142, 155), (142, 157), (142, 158), (142, 160), (142, 162), (142, 163), (142, 165), (142, 167), (143, 15), (143, 17), (143, 18), (143, 20), (143, 22), (143, 23), (143, 25), (143, 27), (143, 28), (143, 30), (143, 32), (143, 33), (143, 35), (143, 37), (143, 38), (143, 40), (143, 42), 
(143, 43), (143, 45), (143, 47), (143, 48), (143, 50), (143, 52), (143, 53), (143, 55), (143, 57), (143, 58), (143, 60), (143, 62), (143, 63), (143, 65), (143, 67), (143, 68), (143, 70), (143, 72), (143, 73), (143, 75), (143, 77), (143, 78), (143, 80), (143, 82), (143, 83), (143, 85), (143, 87), (143, 88), (143, 90), (143, 92), 
(143, 93), (143, 95), (143, 97), (143, 98), (143, 100), (143, 102), (143, 103), (143, 105), (143, 107), (143, 108), (143, 110), (143, 112), (143, 113), (143, 115), (143, 117), (143, 118), (143, 120), (143, 122), (143, 123), (143, 125), (143, 127), (143, 128), (143, 130), (143, 132), (143, 133), (143, 135), (143, 137), (143, 138), (143, 140), (143, 142), 
(143, 143), (143, 145), (143, 147), (143, 148), (143, 150), (143, 152), (143, 153), (143, 155), (143, 157), (143, 158), (143, 160), (143, 162), (143, 163), (143, 165), (143, 167)]